/* library interface for G.722.1, based on: */

/***************************************************************************
**
**   ITU-T G.722.1 (2005-05) - Fixed point implementation for main body and Annex C
**   > Software Release 2.1 (2008-06)
**     (Simple repackaging; no change from 2005-05 Release 2.0 code)
**
**    2004 Polycom, Inc.
**
**   All rights reserved.
**
***************************************************************************/

/***************************************************************************
  Filename:    decode.c    

  Purpose:     Contains the main function for the G.722.1 Annex C decoder
		
  Design Notes:

  WMOPS:     7kHz |    24kbit    |     32kbit
           -------|--------------|----------------
             AVG  |     2.73     |     2.85
           -------|--------------|----------------  
             MAX  |     2.79     |     2.88
           -------|--------------|---------------- 

            14kHz |    24kbit    |     32kbit     |     48kbit
           -------|--------------|----------------|----------------
             AVG  |     5.28     |     5.53       |     5.85   
           -------|--------------|----------------|----------------
             MAX  |     5.56     |     5.77       |     5.95   
           -------|--------------|----------------|----------------

***************************************************************************/
/***************************************************************************
 Include files                                                           
***************************************************************************/
#include <stdio.h>
#include <string.h>
#include "g7221.h"
#include "defs.h"
#include "count.h"

/************************************************************************************
 Local type declarations                                             
*************************************************************************************/

/* Holds all state needed from one frame to the next */
typedef struct
{
    Word16 decoder_mlt_coefs[MAX_DCT_LENGTH];
    Word16 mag_shift;
    Word16 old_mag_shift;
    Word16 old_decoder_mlt_coefs[MAX_DCT_LENGTH];
    Word16 old_samples[MAX_DCT_LENGTH>>1];
    Rand_Obj randobj;
} DECODER_STATE;

/* This object is used to control the decoder configuration */
typedef struct 
{
    Word16  syntax;
    Word32  bit_rate;
    Word16  bandwidth;
    Word16  number_of_bits_per_frame;
    Word16  number_of_regions;
    Word16  frame_size;
} DECODER_CONTROL;

/* The persistant object we hand back to the caller */
/* typedef g7221_handle_s g7221_handle; */
struct g7221_handle_s
{
    DECODER_CONTROL control;
    DECODER_STATE state;
};

/************************************************************************************
 Constant definitions                                                    
*************************************************************************************/
#define MAX_SAMPLE_RATE     32000
#define MAX_FRAMESIZE   (MAX_SAMPLE_RATE/50)

/***************************************************************************
 Local function declarations                                             
***************************************************************************/

__declspec( dllexport )
void g7221_decode_frame(
    g7221_handle    *handle,
    int16_t         *code_words,
    int16_t         *sample_buffer)
{
    Word16 i;
    Bit_Obj bitobj;

    /* reinit the current word to point to the start of the buffer */
    bitobj.code_word_ptr = code_words;
    bitobj.current_word =  *code_words;
    bitobj.code_bit_count = 0;
    bitobj.number_of_bits_left = handle->control.number_of_bits_per_frame;
        
    /* process the out_words into decoder_mlt_coefs */
    decoder(&bitobj,
            &handle->state.randobj,
            handle->control.number_of_regions,
            handle->state.decoder_mlt_coefs,
            &handle->state.mag_shift,
            &handle->state.old_mag_shift,
            handle->state.old_decoder_mlt_coefs,
            0 // no frame error
            );

        
        
    /* convert the decoder_mlt_coefs to samples */
    rmlt_coefs_to_samples(
            handle->state.decoder_mlt_coefs,
            handle->state.old_samples,
            sample_buffer,
            handle->control.frame_size,
            handle->state.mag_shift);

    /* For ITU testing, off the 2 lsbs. */
    for (i=0; i<handle->control.frame_size; i++)
        sample_buffer[i] &= 0xfffc;
}

__declspec( dllexport )
g7221_handle * g7221_init(int bytes_per_frame, int bandwidth)
{
    Word32 bit_rate;

    g7221_handle * handle = malloc(sizeof(g7221_handle));
    if (!handle)
    {
        goto fail;
    }

    memset(handle, 0, sizeof(g7221_handle));

    /* determine format */

    bit_rate = (Word32)bytes_per_frame*8*50;

    handle->control.bit_rate = bit_rate;
    if ((bit_rate < 16000) || (bit_rate >bit_rate > 48000) ||
        ((bit_rate/800)*800 != bit_rate))
    {
        fprintf(stderr,"codec: Error. bit-rate must be multiple of 800 between 16000 and 48000\n");
        goto fail;
    }

    handle->control.number_of_bits_per_frame = bytes_per_frame*8;

    handle->control.bandwidth = bandwidth;
    if (handle->control.bandwidth == 7000)
    {
        handle->control.number_of_regions = NUMBER_OF_REGIONS;
        handle->control.frame_size = MAX_FRAMESIZE >> 1;
    }
    else if (handle->control.bandwidth == 14000)
    {
        handle->control.number_of_regions = MAX_NUMBER_OF_REGIONS;
        handle->control.frame_size = MAX_FRAMESIZE;
    }
    else
    {
        fprintf(stderr,"codec: Error. bandwidth must be 7000 or 14000\n");
        goto fail;
    }

    g7221_reset(handle);

    return handle;

fail:
    if (handle)
    {
        free(handle);
    }
    return NULL;
}

__declspec( dllexport )
void g7221_reset(g7221_handle *handle)
{
    Word16 i;

    memset(&handle->state, 0, sizeof(handle->state));

    handle->state.old_mag_shift = 0;

    /* initialize the coefs history */
    for (i=0;i<handle->control.frame_size;i++)
        handle->state.old_decoder_mlt_coefs[i] = 0;    

    for (i = 0;i < (handle->control.frame_size >> 1);i++)
  	    handle->state.old_samples[i] = 0;
    
    /* initialize the random number generator */
    handle->state.randobj.seed0 = 1;
    handle->state.randobj.seed1 = 1;
    handle->state.randobj.seed2 = 1;
    handle->state.randobj.seed3 = 1;
}

__declspec( dllexport )
void g7221_free(g7221_handle *handle)
{
    if (handle)
    {
        free(handle);
    }
}
